(function(){
 function measureOpticalBounds(canvas) {
     var ctx = canvas.getContext("2d"),
         spaceWidth = ctx.measureText(" ").width,
         cWidth = canvas.width,
         cHeight = canvas.height,
         data = ctx.getImageData(0, 0, cWidth, cHeight).data,
         first = false,
         last = false,
         r = cHeight,
         c = 0;

     // Find the last line with a non-white pixel
     while (!last && r) {
         r--;
         for (c = 0; c < cWidth; c++) {
             if (data[r * cWidth * 4 + c * 4 + 3]) {
                 last = r;
                 break;
             }
         }
     }

     // Find the first line with a non-white pixel
     while (r) {
         r--;
         for (c = 0; c < cWidth; c++) {
             if (data[r * cWidth * 4 + c * 4 + 3]) {
                 first = r;
                 break;
                 }
             }
         }
         return {
             top: first,
             bottom: last,
             spaceWidth: spaceWidth
         };
    };

     var fontLeftPlacement = 100, //?
         fontBaselinePlacement = 300, //?
         measureFontSizeBase = 200, //?
         noTestChars = '‘’“"”`.:,;' + "'",
         defaultData = {
         postScriptName: "",
         family: "",
         weight: "",
         category: "",
         label: "",
         filename: "",
         uppercaseChars: "ABCDEFGHIJKLMNOPQRSTUVWXYZ" +
         "ÅÀÁÃÄÂßÆŒÇÐØÈÉÊËÌÍÎÏÒÓÔÕÖÙÚÛÜÝÑ" +
         "1234567890" +
         '‘’“"”`' +
         "'",
         lowercaseChars: "abcdefghijklmnopqrstuvwxyz @<>-+=.:,~;" +
         "àáâäæãåçèéêëÿûüùúîïíìôöòóœøõñð" +
         "!#$%^&*(){}[]?/_",
         capTest: "X",
         meanTest: "x",
         ascentTest: "b d l",
         descentTest: "p g",
         ascentDescentTest: "Q g b d q",
         tailTest: "Q",
         meanAccentTest: "àá",
         capAccentTest: "ÀÁ",
         spaceWidthRatio: 0,
         capAccentRatio: 0,
         capRatio: 0,
         meanAccentRatio: 0,
         meanRatio: 0,
         meanWidthRatio: 0,
         ascentRatio: 0,
         baselineOffsetRatio: 0,
         iosBaselineOffsetModifier: 0,
         tailRatio: 0,
         descentRatio: 0,
         italicAngle: 0,
         capAccentChars: "",
         capChars: "",
         ascentChars: "",
         meanAccentChars: "",
         meanChars: " ",
         baseChars: "",
         tailChars: "",
         descentChars: "",
         displaySample: "Ag",
         popularity: 5,
         formality: 5,
         blackness: 5,
         drama: 5,
         width: 5,
         allCapsAllowed: true,
         displayAllowed: true,
         bodyAllowed: false,
         smallCaps: false,
         serif: false,
         slab: false,
         handdrawn: false,
         calligraphic: false,
         decorative: false,
         modern: false,
         square: false,
         geometric: false,
         italic: false,
         style: "",
         deprecated: false,
         isPublic: true
         };


 function generateMetrics (canvasFactory, fontName, cssName, fontInfo) {
     fontInfo = fontInfo || {};

     var allTestChars = defaultData.uppercaseChars + defaultData.lowercaseChars,
         derivedMetrics = _.extend({}, defaultData, fontInfo),
         fontLeftPlacement = 100,
         fontBaselinePlacement = 250,
         measureFontSizeBase = 200,
         canvasHeight = 400,
         canvasWidth = 800,
         aboveDivider = 0,
         belowDivider = 0,
         ascentDescentRatio = 0,
         meanWidthRatio = 0,
         spaceWidthRatio = 0,
         xWidthRatio = 0,
         ascentRatio = 0,
         meanAccentRatio = 0,
         capAccentRatio = 0,
         capRatio = 0,
         meanRatio = 0,
         descentRatio = 0,
         tailRatio = 0,
         baselineOffsetRatio = 0,
         key,
         overrides = {},
         charsToForceLowercase = overrides.lowercaseChars || "",
         charsToForceUppercase = overrides.uppercaseChars || "",
         boundsTestCanvas = canvasFactory(canvasWidth, canvasHeight),
         boundsTestCanvasCtx = boundsTestCanvas.getContext("2d");

     function getBoundsOfString(str, baseline) {
         boundsTestCanvasCtx.clearRect(0, 0, canvasWidth, canvasHeight);
         var ctx = boundsTestCanvas.getContext("2d");
         ctx.fillStyle = "#000000";
         ctx.textAlign = "left";

         var fontYPlacement = fontBaselinePlacement;
         if (baseline) {
            ctx.textBaseline = baseline;
             if ( baseline === "middle" ) {
                fontYPlacement = canvasHeight / 2;
             }
         } else {
            ctx.textBaseline = "alphabetic";
         }

         ctx.font = (fontInfo.style || 'normal') + ' ' + (fontInfo.weight || 400) + ' ' + measureFontSizeBase + 'px ' + '"' + cssName + '",serif';
         ctx.fillText(str, fontLeftPlacement, fontYPlacement);

         var bounds = measureOpticalBounds(boundsTestCanvas);
         return bounds;
     }

     for (key in charsToForceLowercase){
         var lowerChar = charsToForceLowercase[key],
         uIndex = derivedMetrics.uppercaseChars.indexOf(lowerChar);
         if (uIndex >= 0) {
             derivedMetrics.uppercaseChars = spliceString(derivedMetrics.uppercaseChars,uIndex,1);
             derivedMetrics.lowercaseChars += lowerChar;
         }
     }

     for (key in charsToForceUppercase) {
         var upperChar = charsToForceUppercase[key],
         lIndex = derivedMetrics.lowercaseChars.indexOf(upperChar);
         if (lIndex >= 0) {
             derivedMetrics.lowercaseChars = spliceString(derivedMetrics.lowercaseChars,lIndex,1);
             derivedMetrics.uppercaseChars += upperChar;
         }
     }
     overrides.uppercaseChars = derivedMetrics.uppercaseChars;
     overrides.lowercaseChars = derivedMetrics.lowercaseChars;

     derivedMetrics = _.merge(derivedMetrics, overrides);

     var boundsCap = getBoundsOfString(derivedMetrics.capTest),
         boundsMean = getBoundsOfString(derivedMetrics.meanTest),
         boundsAscent = getBoundsOfString(derivedMetrics.ascentTest),
         boundsDescent = getBoundsOfString(derivedMetrics.descentTest),
         boundsAscentDescent = getBoundsOfString(derivedMetrics.ascentDescentTest, "middle"),
         boundsTail = getBoundsOfString(derivedMetrics.tailTest),
         boundsMeanAccent = getBoundsOfString(derivedMetrics.meanAccentTest),
         boundsCapAccent = getBoundsOfString(derivedMetrics.capAccentTest);

     // derived values from measured bounds
     aboveDivider = (canvasHeight / 2) - boundsAscentDescent.top;
     belowDivider = boundsAscentDescent.bottom - (canvasHeight / 2);
     ascentDescentRatio = (boundsAscentDescent.bottom - boundsAscentDescent.top) / measureFontSizeBase;
     spaceWidthRatio = Math.round((boundsCap.spaceWidth / measureFontSizeBase) * 1000) / 1000;
     xWidthRatio = Math.round((boundsMean.xWidth / measureFontSizeBase) * 1000) / 1000;
     ascentRatio = (boundsMean.bottom - boundsAscent.top) / measureFontSizeBase;
     meanAccentRatio = (boundsMean.bottom - boundsMeanAccent.top) / measureFontSizeBase;
     capAccentRatio = (boundsMean.bottom - boundsCapAccent.top) / measureFontSizeBase;
     descentRatio = (boundsDescent.bottom - boundsMean.bottom) / measureFontSizeBase;
     tailRatio = (boundsTail.bottom - boundsMean.bottom) / measureFontSizeBase;
     capRatio = (boundsMean.bottom - boundsCap.top) / measureFontSizeBase;
     meanRatio = (boundsMean.bottom - boundsMean.top) / measureFontSizeBase;
     meanWidthRatio = boundsMean.spaceWidth / measureFontSizeBase;
     baselineOffsetRatio = (fontBaselinePlacement - boundsMean.bottom) / measureFontSizeBase;

     if (_.isNumber(overrides.baselineOffsetRatio)){
         var offsetPx = overrides.baselineOffsetRatio * measureFontSizeBase,
         tempBotX = fontBaselinePlacement - offsetPx;
         ascentRatio = (tempBotX - boundsAscent.top) / measureFontSizeBase;
         meanAccentRatio = (tempBotX - boundsMeanAccent.top) / measureFontSizeBase;
         capAccentRatio = (tempBotX - boundsCapAccent.top) / measureFontSizeBase;
         descentRatio = (boundsDescent.bottom - tempBotX) / measureFontSizeBase;
         tailRatio = (boundsTail.bottom - tempBotX) / measureFontSizeBase;
         capRatio = (tempBotX - boundsCap.top) / measureFontSizeBase;
         meanRatio = (tempBotX - boundsMean.top) / measureFontSizeBase;
     }

     var _aboveDivider = _.isNumber(overrides.aboveDivider) ? overrides.aboveDivider : aboveDivider,
     _belowDivider = _.isNumber(overrides.belowDivider) ? overrides.belowDivider : belowDivider,
     _ascentDescentRatio = _.isNumber(overrides.ascentDescentRatio) ? overrides.ascentDescentRatio : ascentDescentRatio,
     _meanWidthRatio = _.isNumber(overrides.meanWidthRatio) ? overrides.meanWidthRatio : meanWidthRatio,
     _spaceWidthRatio = _.isNumber(overrides.spaceWidthRatio) ? overrides.spaceWidthRatio : spaceWidthRatio,
     _xWidthRatio = _.isNumber(overrides.xWidthRatio) ? overrides.xWidthRatio : xWidthRatio,
     _ascentRatio = _.isNumber(overrides.ascentRatio) ? overrides.ascentRatio : ascentRatio,
     _meanAccentRatio = _.isNumber(overrides.meanAccentRatio) ? overrides.meanAccentRatio : meanAccentRatio,
     _capAccentRatio = _.isNumber(overrides.capAccentRatio) ? overrides.capAccentRatio : capAccentRatio,
     _descentRatio = _.isNumber(overrides.descentRatio) ? overrides.descentRatio : descentRatio,
     _tailRatio = _.isNumber(overrides.tailRatio) ? overrides.tailRatio : tailRatio,
     _capRatio = _.isNumber(overrides.capRatio) ? overrides.capRatio : capRatio,
     _meanRatio = _.isNumber(overrides.meanRatio) ? overrides.meanRatio : meanRatio,
     _baselineOffsetRatio = _.isNumber(overrides.baselineOffsetRatio) ? overrides.baselineOffsetRatio : baselineOffsetRatio,
     baseLine = fontBaselinePlacement - _baselineOffsetRatio * measureFontSizeBase,
     ascentLine = baseLine - _ascentRatio * measureFontSizeBase,
     meanAccentLine = baseLine - _meanAccentRatio * measureFontSizeBase,
     capAccentLine = baseLine - _capAccentRatio * measureFontSizeBase,
     descentLine = baseLine + _descentRatio * measureFontSizeBase,
     tailLine = baseLine + _tailRatio * measureFontSizeBase,
     capLine = baseLine - _capRatio * measureFontSizeBase,
     meanLine = baseLine - _meanRatio * measureFontSizeBase;

     derivedMetrics.aboveDivider = _aboveDivider;
     derivedMetrics.belowDivider = _belowDivider;
     derivedMetrics.ascentDescentRatio = _ascentDescentRatio;
     derivedMetrics.meanWidthRatio = _meanWidthRatio;
     derivedMetrics.spaceWidthRatio = _spaceWidthRatio;
     derivedMetrics.xWidthRatio = _xWidthRatio;
     derivedMetrics.ascentRatio = _ascentRatio;
     derivedMetrics.meanAccentRatio = _meanAccentRatio;
     derivedMetrics.capAccentRatio = _capAccentRatio;
     derivedMetrics.descentRatio = _descentRatio;
     derivedMetrics.tailRatio = _tailRatio;
     derivedMetrics.capRatio = _capRatio;
     derivedMetrics.capRatio = _capRatio;
     derivedMetrics.meanRatio = _meanRatio;
     derivedMetrics.baselineOffsetRatio = _baselineOffsetRatio;

     var tailOffsetFromBase = Math.abs(tailLine - baseLine),
        hasTail = tailOffsetFromBase > 5;

     for (key in allTestChars){
         var char = allTestChars[key],
             isUppercase = derivedMetrics.uppercaseChars.indexOf(char) >= 0,
             charBounds = getBoundsOfString(char),
             capDiff = Math.abs(charBounds.top - capLine),
             capAccentDiff = Math.abs(charBounds.top - capAccentLine),
             meanAccentDiff = Math.abs(charBounds.top - meanAccentLine),
             meanDiff = Math.abs(charBounds.top - meanLine),
             ascentDiff = Math.abs(charBounds.top - ascentLine),
             baseDiff = Math.abs(Math.max(0,charBounds.bottom - baseLine)),
             tailDiff = Math.abs(charBounds.bottom - tailLine),
             descentDiff = Math.abs(charBounds.bottom - descentLine),
             topBoundsKey = isUppercase ? "capChars" : "meanChars",
             bottomBoundsKey = "baseChars",
             skipOpticalTest = noTestChars.indexOf(char) >= 0;

         if (!skipOpticalTest) {
             var curTopDiff = meanDiff,
                curBotDiff = baseDiff;

             if (meanAccentDiff < curTopDiff * 1.2) {
                 curTopDiff = meanAccentDiff;
                 topBoundsKey = "meanAccentChars";
             }
             if (capDiff < curTopDiff * 1.2) {
                 curTopDiff = capDiff;
                 topBoundsKey = "capChars";
             }
             if (ascentDiff < curTopDiff * 1.2) {
                 curTopDiff = ascentDiff;
                 topBoundsKey = "ascentChars";
             }
             if (capAccentDiff < curTopDiff * 1.2) {
                 curTopDiff = capAccentDiff;
                 topBoundsKey = "capAccentChars";
             }
             if (tailDiff < curBotDiff * 1.2){
                 curBotDiff = tailDiff;
                 bottomBoundsKey = "tailChars";
             }
             if (descentDiff < curBotDiff * 1.2){
                 curBotDiff = descentDiff;
                 bottomBoundsKey = "descentChars";
             }
             if (overrides.meanChars && overrides.meanChars.indexOf(char) >= 0){
                topBoundsKey = "meanChars";
             }
             if (overrides.baseChars && overrides.baseChars.indexOf(char) >= 0){
                bottomBoundsKey = "baseChars";
             }
             if (overrides.capChars && overrides.capChars.indexOf(char) >= 0){
                topBoundsKey = "capChars";
             }
             if (overrides.meanAccentChars && overrides.meanAccentChars.indexOf(char) >= 0){
                topBoundsKey = "meanAccentChars";
             }
             if (overrides.ascentChars && overrides.ascentChars.indexOf(char) >= 0){
                topBoundsKey = "ascentChars";
             }
             if (overrides.capAccentChars && overrides.capAccentChars.indexOf(char) >= 0){
                topBoundsKey = "capAccentChars";
             }
             if (overrides.tailChars && overrides.tailChars.indexOf(char) >= 0){
                bottomBoundsKey = "tailChars";
             }
             if (overrides.descentChars && overrides.descentChars.indexOf(char) >= 0){
                bottomBoundsKey = "descentChars";
             }
         }
         derivedMetrics[topBoundsKey] += char;
         derivedMetrics[bottomBoundsKey] += char;
     }

     derivedMetrics.label = _.trim(fontName);
     derivedMetrics.postScriptName = _.trim(fontInfo.postScriptName || fontName);
     derivedMetrics.family = _.trim(fontName);

     delete derivedMetrics.capTest;
     delete derivedMetrics.meanTest;
     delete derivedMetrics.ascentTest;
     delete derivedMetrics.descentTest;
     delete derivedMetrics.tailTest;
     delete derivedMetrics.meanAccentTest;
     delete derivedMetrics.capAccentTest;

     return derivedMetrics;
 }

 var getVariation = function(tkFontId, callback) {
     var xmlHttp = new XMLHttpRequest();
     var url = "https://cctypekit.adobe.io/v1/variations/" + tkFontId;
     xmlHttp.open("GET", url, true);
     xmlHttp.onload = function(e) {
     if(xmlHttp.readyState === 4) {
     if(xmlHttp.status === 200) {
     var jsonResponse = JSON.parse(xmlHttp.responseText);
     callback(jsonResponse, null);
     }else{
     callback(null, xmlHttp.statusText)
     console.error(xmlHttp.statusText);
     }
     }
     };
     xmlHttp.onerror = function(e) {
     console.error(xmlHttp.statusText);
     callback(null, xmlHttp.statusText);
     };
     xmlHttp.setRequestHeader("X-Api-Key", "TypekitSwaggerUI");
     xmlHttp.send(null);

 }

 window.getFontMetricsByTKFontId = function(tkFontId) {
    getVariation(tkFontId, function(variationsData, error) {
        console.log(variationsData);
        var webFontId = variationsData.family.web_id;
        var variations = [variationsData.font.web.fvd];
//        var cssName = variationsData.family.css_font_stack;
        var cssName = variationsData.family.name;
        var style = variationsData.style;
        var weight = variationsData.font.web.weight;
        var postScriptName = variationsData.postscript_name;
        var fullName = variationsData.name;
        calculateFontMetrics(tkFontId, webFontId, variations, cssName, style, weight, postScriptName, fullName, variationsData);
    });
 }


 calculateFontMetrics = function(tkFontId, webFontId, variations, cssName, style, weight, postScriptName, fullName, variationsData) {
     var fontInfo = {
        style:style,
        weight:weight,
        postScriptName:postScriptName,
        fullName:fullName,
        tkFontId: tkFontId,
        cssName: cssName,
        familyName:variationsData.family.name || "",
        preferredFamilyName: variationsData.preferred_family_name || "",
        preferredSubfamilyName: variationsData.preferred_subfamily_name || ""
     }

     var typekitInfo = {
         id:webFontId,
         variations:variations,
         css_name:cssName
     }

    TypekitPreview.setup({
                          'auth_id': 'slte',
                          'auth_token': 'Hik3Pm6JnhEwLxQ0zehfgHbPDnVkFnfVeM5Dl15HLM9Nvtt1z6Zcx%2FCrp%2Ba6IPG%2F4rar1RfQpMk18PnaqVj5UR3JusyvaDexkb4zuYwE6edbEtYVoeY8I0RGujcBA982%2FgXgDjNyLWrd6f8i%2FsbDs1mQLbIorXyId0bC0yU2kZo%3D',
                          });

    TypekitPreview.load([typekitInfo], {
                         // Font event callbacks

                         loading: function () {
                            // Called when these fonts are requested
                         },
                         active: function () {
                             // Called after these fonts are done loading and at least one is active
                             var testMatrix = generateMetrics(
                                                              function(width, height) {
                                                              canvas = document.createElement('canvas');
                                                              canvas.width = width;
                                                              canvas.height = height;
                                                              return canvas;
                                                              },
                                                              fontInfo.fullName, cssName, fontInfo
                                                              )

                             console.log(JSON.stringify(testMatrix, null, 4));
                             TypekitMetricsJavaScriptInterface.successHandler(JSON.stringify(testMatrix, null, 4));
                         },
                         inactive: function () {
                             // Called after these fonts are done loading and none are active
                             // Also called immediately if fonts are not supported
                            TypekitMetricsJavaScriptInterface.failHandler("inactive font");

                         },
                         fontloading: function (fontFamily, fontDescription) {
                            // Called as each font is requested
                         },
                         fontactive: function (fontFamily, fontDescription) {
                            // Called after a font finishes loading and is active
                         },
                         fontinactive: function (fontFamily, fontDescription) {
                            // Called after a font fails to load
                            console.log("font fails to load fontFamily: " + fontFamily + " ,fontDescript: " + fontDescription);
                            TypekitMetricsJavaScriptInterface.failHandler("font fails to load fontFamily: " + fontFamily + " ,fontDescript: " + fontDescription);
                         }
                        });
    }
 })();
